from typing import Any
import zlib
import asyncio
from uuid import UUID
import logging

from pproto_py.core import Formats, Commands, FormatsException
from pproto_py.schemas import BaseMessage, Compression, Status
from pproto_py.base import Base
from pproto_py.client.listener import MessageReceiver
from pproto_py.client.message_pool import MessagePool

logger = logging.getLogger(__name__)


class Client(Base):
    def __init__(
        self,
        host="127.0.0.1",
        port=8888,
        *,
        format=Formats.JSON_PROTOCOL_FORMAT.value,
        compatible=Commands.Compatible.value,
        use_compress=False,
        compress_level: int = Compression.UNDEFINED,
    ):
        self.__host = host
        self.__port = port
        self.writer, self.reader = None, None  # type: asyncio.StreamWriter | None, asyncio.StreamReader | None
        self.__format = format
        self.compatible = compatible
        self.use_compress = use_compress
        self.compress_level = compress_level
        #
        self.__message_receiver: MessageReceiver | None = None
        self.__message_pool: MessagePool | None = None
        self.__message_receiver_task: asyncio.Task | None = None

    @classmethod
    async def create_connect(cls, host: str, port: int, *args: Any, **kwargs: Any) -> "Client":
        self = cls(host, port, *args, **kwargs)
        await self.connect()
        return self

    @property
    def is_connected(self) -> bool:
        return self.writer is not None

    def __str__(self):
        return f"<{self.__class__.__name__}>({self.__host}:{self.__port})"

    async def connect(self) -> bool:
        logger.debug(f"Connected {self}")
        reader, writer = await asyncio.open_connection(self.__host, self.__port)
        self.writer = writer
        self.reader = reader
        # init message utils
        self.__message_receiver = MessageReceiver(reader)
        self.__message_pool = MessagePool()
        await self.hello_message()
        await self.compatible_message()
        # Start task
        self.__message_receiver_task = asyncio.create_task(
            self.__message_listener(self.__message_receiver, self.__message_pool)
        )
        return True

    async def hello_message(self) -> None:
        format = UUID(self.__format).bytes
        self.writer.write(format)
        await self.writer.drain()
        data = await self.reader.read(16)
        if data != format:
            raise FormatsException(error="The server format is different from the client")

    async def compatible_message(self) -> None:
        data_size = int.from_bytes(await self.reader.read(4))
        await self.reader.read(data_size)
        message = BaseMessage(
            command=self.compatible,
            maxTimeLife=5,
        )
        self.writer.write(self.swap32_len(message=message))
        await self.writer.drain()
        self.writer.write(message.model_dump_json().encode())
        await self.writer.drain()
        # TODO data_compatible checking

    @staticmethod
    async def __case_message(message: BaseMessage, callback_func: dict) -> None:
        match message.flags.exec_status.value:
            case Status.SUCCESS.value:
                await callback_func["answer"](message.content)
            case Status.FAILED.value:
                await callback_func["failed"](message.content)
            case Status.ERROR.value:
                await callback_func["error"](message.content)
            case Status.UNKNOWN.value:
                await callback_func["unknown"](message.content)

    async def __message_listener(self, message_receiver: MessageReceiver, message_pool: MessagePool):
        # TODO::Heartbeat or pinp-pong:@cans1194
        base_commands_handlers = {
            Commands.CloseConnection.value: self.close(),
        }
        try:
            async for message in message_receiver:  # type: BaseMessage
                logger.debug(f"Received message: {message!r} on {self!r}")
                #
                handler = base_commands_handlers.get(message.command, None)
                if handler:
                    if asyncio.iscoroutinefunction(handler):
                        await handler()
                else:
                    message_pool.add_message(message)
        except asyncio.CancelledError:
            logger.error(f"Connection {self!r} canceled")
            self.__message_receiver_task.cancel()
            raise

    async def _read_with_callback(self, message: BaseMessage, callback_func: dict) -> None:
        received_message = await self.__message_pool.get_message(message.id, message.max_time_life)
        await self.__case_message(received_message, callback_func)
        self.__message_pool.dell_message(received_message)

    async def write(self, message: BaseMessage) -> None:
        # TODO переписат ьс првоеркой на use_commpress
        self.writer.write(self.swap32_len(message=message, compress=self.use_compress))
        await self.writer.drain()
        if message.flags.compression.value == Compression.DISABLE.value:
            self.writer.write((message.model_dump_json()).encode())
        if self.use_compress:
            header = len(message.model_dump_json().encode("utf-8")).to_bytes(4, byteorder="big")
            data = zlib.compress(message.model_dump_json().encode("utf-8"))
            self.writer.write(header + data)
        await self.writer.drain()

    async def write_with_callback(self, message: BaseMessage, callback_func: dict) -> None:
        self.writer.write(self.swap32_len(message=message))
        await self.writer.drain()
        if message.flags.compression.value == Compression.DISABLE.value:
            self.writer.write((message.model_dump_json()).encode())
        if message.flags.compression.value == Compression.NONE.value:
            self.writer.write(zlib.compress(message.model_dump_json().encode("utf-8")))
        await self.writer.drain()
        await self._read_with_callback(message, callback_func)

    async def close(self) -> None:
        self.writer.close()
        await self.writer.wait_closed()



from typing import TypeVar, Type

from pydantic import BaseModel, TypeAdapter

from pproto_py.client import Client
from pproto_py.content import BaseContent


ModelType = TypeVar("ModelType", bound=BaseModel)
ContentType = TypeVar("ContentType", bound=BaseContent)


def session(func):
    session = Client()

    def wrapper(*args, **kwargs):
        return func(*args, session, **kwargs)

    return wrapper


async def format_answer(raw_records: dict, model: BaseModel) -> BaseModel | None:
    if not raw_records:
        return None
    return map(lambda x: model(**x), raw_records)


def to_model(model: Type[ModelType | ContentType]):
    def outer(func):
        async def inner(*args, **kwargs):
            content_model: BaseModel = TypeAdapter(model).validate_python(args[1])
            if len(args[2:]) != 0:
                new_args = (args[0], content_model, args[2:])
            else:
                new_args = (args[0], content_model)
            res = await func(*new_args, **kwargs)
            return res

        return inner

    return outer



import zlib
from pproto_py.schemas import BaseMessage


class Base:
    @staticmethod
    def swap32_len(message: BaseMessage, compress: bool = False) -> bytes:
        if compress:
            message_size = len(zlib.compress(message.model_dump_json().encode("utf-8"))) * -1
        else:
            message_size = len(message.model_dump_json().encode("utf-8"))
        bytes_size = int.from_bytes(
            message_size.to_bytes(4, byteorder="little", signed=True), byteorder="big", signed=False
        ).to_bytes(4, byteorder="little")
        return bytes_size



class PprotoCommonException(Exception):
    def __init__(self, error: str) -> None:
        super().__init__()
        self.error = error


class FormatsException(PprotoCommonException):
    def __init__(self, error: str) -> None:
        super().__init__(error=error)
        self.error = error


class CompatibleException(PprotoCommonException):
    def __init__(self, code: int, error: str) -> None:
        super().__init__()
        self.error = error


class TypeMessageError(PprotoCommonException):
    def __init__(self, error: str) -> None:
        super().__init__()
        self.error = error


class PprotoConnectionError(PprotoCommonException):
    ...


class CommandTimeLifeOutError(PprotoCommonException):
    ...



import asyncio
import zlib
from collections.abc import AsyncIterable

from pproto_py.schemas import BaseMessage
from pproto_py.core.exceptions import PprotoConnectionError


class MessageReceiver(AsyncIterable):
    def __init__(self, reader: asyncio.StreamReader):
        self.reader = reader
        self.lock = asyncio.Lock()

    def __aiter__(self) -> "MessageReceiver":
        return self

    async def __anext__(self) -> BaseMessage:
        return await self.get_message()

    async def get_message(self) -> BaseMessage:
        async with self.lock:
            if self.reader is None:
                raise PprotoConnectionError("Connection closed")
            try:
                data_size = int.from_bytes(await self.reader.read(4), signed=True)
                data = await self.reader.readexactly(abs(data_size))
                if data_size < 0:
                    data = zlib.decompress(data[4:])
                message = BaseMessage.from_json(data)
            except asyncio.IncompleteReadError as err:
                raise PprotoConnectionError("") from err
            except ConnectionRefusedError as err:
                raise PprotoConnectionError("") from err
            except ConnectionResetError as err:
                raise PprotoConnectionError("") from err
            except ConnectionError as err:
                raise PprotoConnectionError("") from err
            except OSError as err:
                raise PprotoConnectionError(f"Communication error: {err!r}") from err
        return message



from typing import Callable, Optional, Dict
from uuid import UUID
from pproto_py.content import BaseContent
from pproto_py.schemas import BaseMessage


class Router:
    def __init__(
        self,
    ) -> None:
        self.routers: Dict[UUID, Command] = {}

    async def event(id: UUID, **kawrgs):
        pass

    async def command(id: UUID, response_model: BaseContent, **kwargs):
        pass


class Command:
    def __init__(
        self,
        func: Callable,
        response_model=Optional[BaseContent | BaseMessage],
    ):
        self.func = func
        self.response_model = response_model



from .schemas import (
    Priority,
    Status,
    Type,
    BaseMessage,
    Compression,
    Serialize,
    FlagField,
    ByteField,
    FlagMessage,
    BasePprotoErrorContent,
)
from .client import Client
from .content import BaseContent, format_answer, to_model
from .server import Pproto

__all__ = [
    "Priority",
    "Status",
    "Type",
    "BaseMessage",
    "Compression",
    "Serialize",
    "BaseContent",
    "Client",
    "FlagField",
    "ByteField",
    "FlagMessage",
    "format_answer",
    "to_model",
    "Pproto",
    "BasePprotoErrorContent",
]



from .message import Priority, Status, Type, BaseMessage, Compression, Serialize, FlagField, ByteField, FlagMessage
from .exceptions import BasePprotoErrorContent

__all__ = [
    "Priority",
    "Status",
    "Type",
    "BaseMessage",
    "Compression",
    "Serialize",
    "FlagField",
    "ByteField",
    "FlagMessage",
    "BasePprotoErrorContent",
]



import asyncio
from typing import TypeVar

from pproto_py.schemas import BaseMessage
from pproto_py.core.exceptions import CommandTimeLifeOutError

MessageKeyT = TypeVar("MessageKeyT")


class MessagePool(dict):
    _pending: dict[MessageKeyT, asyncio.Event] = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            pass
        try:
            event = self._pending[key]
        except KeyError:
            self._pending[key] = asyncio.Event()
            event = self._pending[key]
        await event.wait()
        return super().__getitem__(key)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if key in self._pending:
            self._pending.pop(key).set()

    def add_message(self, message: BaseMessage):
        self[message.id] = message

    def dell_message(self, message: BaseMessage):
        if (message_id := message.id) in self:
            super().__delitem__(message_id)
        else:
            raise KeyError("Message does not exist in MessagePool")

    async def get_message(self, message_id: MessageKeyT, timeout=None) -> BaseMessage:
        try:
            async with asyncio.timeout(timeout if timeout != -1 else None):
                return await self.__getitem__(message_id)
        except TimeoutError as err:
            raise CommandTimeLifeOutError(f"Command life-time out {timeout}: {err!r}") from err



from uuid import UUID
from pydantic import BaseModel

from pproto_py.schemas import BaseMessage, Type
from pproto_py.client import Client
from pproto_py.core.exceptions import TypeMessageError


class BaseContent(BaseModel):
    async def send(self, server: Client, **message_params) -> UUID:
        message = BaseMessage(**message_params, content=self)
        callable_func = {"answer": self.answer, "error": self.error, "failed": self.failed, "unknown": self.unknown}
        if message.flags.type.value != Type.EVENT.value:
            await server.write_with_callback(message, callable_func)
        else:
            await server.write(message)
        return message.id

    async def answer() -> None:
        pass

    async def failed() -> None:
        pass

    async def error() -> None:
        pass

    async def unknown() -> None:
        pass

    async def send_wa(self, server: Client, **message_params):
        message = BaseMessage(**message_params, content=self)
        if message.flags.type.value == Type.EVENT:
            raise TypeMessageError("Event don't have a answer!")
        pass



import traceback
import zlib
from uuid import UUID, uuid4
from typing import Dict, Callable
from asyncio.streams import StreamReader, StreamWriter
from pydantic_core import from_json
import asyncio

from pproto_py.core import Formats, Commands, PprotoCommonException, FormatsException, logger
from pproto_py.schemas import BaseMessage, FlagMessage, Type, Status, Compression
from pproto_py.content import BaseContent
from pproto_py.base import Base
from .router import Command


class Pproto(asyncio.Protocol, Base):
    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 8888,
        format=Formats.JSON_PROTOCOL_FORMAT.value,
        compatible=Commands.Compatible.value,
        use_compress=False,
    ):
        self.__host: str = host
        self.__port: int = port
        self.__format = format
        self.compatible = compatible
        self.routers: Dict[UUID, Command] = {}
        self.exception_handlers: Dict[Exception, Callable] = {}
        # self.middleware_routers: Dict[] = {}
        self.clients: Dict[UUID, (StreamReader, StreamWriter)] = {}
        self.use_compress = use_compress

    async def compatible_message(self, user_id: UUID) -> None:
        __reader, __writer = self.clients[user_id]
        message = BaseMessage(
            command=self.compatible,
            maxTimeLife=5,
        )
        __writer.write(self.swap32_len(message=message))
        await __writer.drain()
        __writer.write(message.model_dump_json().encode())
        await __writer.drain()
        data_size = int.from_bytes(await __reader.read(4), signed=True)
        await __reader.read(data_size)
        # TODO check compatible

    async def hello_message(self, user_id: UUID) -> None:
        format = UUID(self.__format).bytes
        __reader, __writer = self.clients[user_id]
        __writer.write(format)
        await __writer.drain()
        data = await __reader.read(16)
        if data != format:
            raise FormatsException(error="The server format is different from the client")

    async def __handle_coroutine(self, reader: StreamReader, writer: StreamWriter):
        await logger.info("INFO -- New client connected")
        try:
            user_id = uuid4()
            self.clients[user_id] = (reader, writer)
            await logger.info("INFO -- Hello message")
            await self.hello_message(user_id)
            await logger.info("INFO -- Compatible message")
            await self.compatible_message(user_id)
            message: BaseMessage = BaseMessage(command=self.compatible)
        except PprotoCommonException as err:
            await logger.error(err)
        while message.command != Commands.CloseConnection.value:
            message_as_dict = await self.read_message(reader)
            await logger.info("INFO -- New message")
            await logger.info(f"INFO -- Message: {str(message_as_dict)}")
            try:
                answer_response = await self.routers[UUID(message_as_dict["command"])].func(message_as_dict)
                answer = await self.make_message(answer_response, BaseMessage(**message_as_dict))
                if answer_response is not None:
                    await self.send_answer(answer, writer)
            except Exception as ex:
                if type(ex) in self.exception_handlers:
                    exception_response = await self.exception_handlers[type(ex)](message_as_dict, ex)
                    await self.send_answer(exception_response, writer)
                else:
                    traceback_text = traceback.format_exc()
                    await logger.error(f"ERROR -- {traceback_text}")
                    raise ex
        await writer.wait_closed()

    async def run(self):
        server = await asyncio.start_server(self.__handle_coroutine, self.__host, self.__port)
        addrs = ", ".join(str(sock.getsockname()) for sock in server.sockets)
        await logger.info(f"INFO -- Serving on {addrs}")
        async with server:
            await server.serve_forever()

    def include_router(self, router) -> None:
        self.routers.apeend(router)

    async def read_message(self, reader) -> BaseMessage:
        data_size = int.from_bytes(await reader.read(4), signed=True)
        data = await reader.read(abs(data_size))
        if data_size < 0:
            data = zlib.decompress(data[4:])
        as_dict = from_json(data.decode("utf-8"))
        as_dict["flags"] = FlagMessage.parse_obj(as_dict["flags"])
        return as_dict

    async def make_message(self, answer: BaseContent, message: BaseMessage) -> BaseMessage:
        message.flags.type.value = Type.ANSWER.value
        message.flags.exec_status.value = Status.SUCCESS.value
        # TODO проверку
        message.content = answer
        return message

    async def send_answer(self, message: BaseMessage, writer: StreamWriter) -> None:
        writer.write(self.swap32_len(message=message, compress=self.use_compress))
        await writer.drain()
        if message.flags.compression.value == Compression.DISABLE.value:
            writer.write((message.model_dump_json()).encode())
        if self.use_compress:
            header = len(message.model_dump_json().encode("utf-8")).to_bytes(4, byteorder="big")
            data = zlib.compress(message.model_dump_json().encode("utf-8"))
            writer.write(header + data)
        await writer.drain()

    def add_pproto_route(
        self,
        id: UUID,
        func: Callable,
        response_model: BaseMessage = None,
    ):
        command = Command(func, response_model)
        self.routers[id] = command

    def command(
        self,
        id: UUID,
        response_model: BaseMessage = None,
    ):
        def decorator(func: Callable) -> Callable:
            self.add_pproto_route(
                id=id,
                func=func,
                response_model=response_model,
            )
            return func

        return decorator

    def event(
        self,
        id: UUID,
    ):
        def decorator(func: Callable) -> Callable:
            self.add_pproto_route(
                id=id,
                func=func,
                response_model=None,
            )
            return func

        return decorator

    def add_exception_handler(self, exc_class: Exception, handler) -> None:  # pragma: no cover
        self.exception_handlers[exc_class] = handler

    def exception_handler(self, exc_class: Exception):
        def decorator(func: Callable) -> Callable:
            self.add_exception_handler(exc_class, func)
            return func

        return decorator

    def middleware(self, middleware_type):
        pass

    def add_middleware():
        pass



from typing import TypeVar
from server import Server
from client import Client

AppType = TypeVar("AppType", bound="PProto")


class PProto:
    def ___init__(
        self: AppType, connection_host: str, connection_port: int, self_host: str, self_port: int
    ) -> "PProto":
        self.__server = Server(self_host, self_port)
        self.__client = Client(connection_host, connection_port)

    def inlucde_router():
        pass



from setuptools import setup, find_packages


with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="pproto-py",
    version="2.1.4a1",
    description='pproto_py is Python implementation of "Point Of View" communication protocol',
    url="https://github.com/TochkaAI/pproto_py-2.0",
    packages=find_packages(),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from enum import Enum
from typing import Annotated
from uuid import UUID, uuid4

from pydantic_core import from_json
from pydantic import BaseModel, Field, model_serializer, PlainSerializer, ConfigDict, TypeAdapter
from pydantic.alias_generators import to_camel


class Type(Enum):
    UNKNOWN = 0
    COMMAND = 1
    ANSWER = 2
    EVENT = 3


class Status(Enum):
    UNKNOWN = 0
    SUCCESS = 1
    FAILED = 2
    ERROR = 3


class Priority(Enum):
    HIGH = 0
    NORMAL = 1
    LOW = 2


class Compression(Enum):
    UNDEFINED = -1
    NONE = 0
    ZIP = 1
    LZMA = 2
    PPMD = 3
    DISABLE = 7


class SerializationFormat:
    BProto = 0  # Бинарный формат
    Json = 1  # Json формат
    # LastFormat = 7 Предполагается, что будет не больше 8 форматов


def int2bit(x: int, size: int):
    """
    Перевод в int в фиксированные 4 байта
    """
    return bin(x)[2:].zfill(size)


class FlagField(BaseModel):
    name: str = Field(...)
    value: int
    size: int

    @model_serializer()
    def serialize_model(self):
        return int2bit(self.value, self.size)


ByteField = Annotated[FlagField, PlainSerializer(lambda x: int2bit(x.value))]


class FlagMessage(BaseModel):
    # 1 Байт
    type: ByteField = ByteField(name="Type", value=Type.COMMAND.value, size=3)
    exec_status: ByteField = ByteField(name="ExecStatus", value=Status.UNKNOWN.value, size=3)
    priority: ByteField = ByteField(name="priority", value=Priority.NORMAL.value, size=2)
    # 2 Байт
    compression: ByteField = ByteField(name="compression", value=Compression.DISABLE.value, size=3)
    tags_is_empty: ByteField = ByteField(name="tagsIsEmpty", value=0, size=1)
    max_time_life_is_empty: ByteField = ByteField(name="maxTimeLife", value=0, size=1)
    content_is_empty: ByteField = ByteField(name="contentIsEmpty", value=1, size=1)
    reserved_2: ByteField = ByteField(name="reserved2", value=0, size=2)
    # 3 Байт
    reserved_3: ByteField = ByteField(name="reserved3", value=0, size=8)
    # 4 Байт
    content_format: ByteField = ByteField(name="contentFormat", value=SerializationFormat.Json, size=3)
    reserved_4: ByteField = ByteField(name="reserved4", value=0, size=4)
    flags_2_is_empty: ByteField = ByteField(name="flags2IsEmpty", value=0, size=1)

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    @model_serializer()
    def serialize_model(self) -> int:
        res = ""
        for k in reversed(self.model_fields.keys()):
            res += getattr(self, k).model_dump()
        return int(res, 2)

    @classmethod
    def parse_obj(cls, flags: int) -> "FlagMessage":
        b_string = bin(flags)[2:].zfill(8 * 4)
        flag = FlagMessage()
        for f in reversed(cls.model_fields.keys()):
            f_field: FlagField = getattr(flag, f)
            f_field.value = int(b_string[0 : f_field.size], 2)
            b_string = b_string[f_field.size :]
        return flag


class Serialize(Enum):
    JSON = "fea6b958-dafb-4f5c-b620-fe0aafbd47e2"


class BaseMessage(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    command: UUID
    flags: FlagMessage = Field(..., default_factory=FlagMessage)
    tags: list = None
    content: BaseModel | dict | None = None
    max_time_life: int = 30

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    @model_serializer()
    def serialize_model(self):
        if self.content is None:
            return {
                "id": self.id,
                "command": self.command,
                "flags": self.flags.model_dump(),
                "maxTimeLife": self.max_time_life,
            }
        else:
            return {
                "id": self.id,
                "command": self.command,
                "flags": self.flags.model_dump(),
                "content": self.content.model_dump(),
                "maxTimeLife": self.max_time_life,
            }

    @classmethod
    def from_json(cls, json: str | bytes | bytearray) -> "BaseMessage":
        # as_dict = ast.literal_eval(data.decode("utf-8"))
        # TODO::Dont' use ast. JSON not equal Python dict.
        """
        ast.literal_eval:
          Safely evaluate an expression node or a Unicode or Latin-1 encoded string containing a Python expression.
          The string or node provided may only consist of the following Python literal structures:
          strings, numbers, tuples, lists, dicts, booleans, and None.

        JSON booleans != Python booleans -> false != False
        JSON null != Python None
        Please read JSON standard ECMA-404
        https://www.json.org
        """
        json_dict = from_json(json)
        json_dict["flags"] = FlagMessage.parse_obj(json_dict["flags"])
        message = TypeAdapter(BaseMessage).validate_python(json_dict)  # type: ignore
        message.content = json_dict["content"]
        return message



from .exceptions import FormatsException, PprotoCommonException
from .commands import Commands, Formats
from .logger import logger

__all__ = [
    "PprotoCommonException",
    "FormatsException",
    "Commands",
    "Formats",
    "logger",
]



from enum import Enum


class Commands(Enum):
    Compatible = "173cbbeb-1d81-4e01-bf3c-5d06f9c878c3"
    Unknown = "4aef29d6-5b1a-4323-8655-ef0d4f1bb79d"
    Error = "b18b98cc-b026-4bfe-8e33-e7afebfbe78b"
    CloseConnection = "e71921fd-e5b3-4f9b-8be7-283e8bb2a531"
    EchoConnection = "db702b07-7f5a-403f-963a-ec50d41c7305"


class Formats(Enum):
    JSON_PROTOCOL_FORMAT = "fea6b958-dafb-4f5c-b620-fe0aafbd47e2"
